#pragma once

#include "ggml.h"
#include "ggml-backend.h"

#ifdef GGML_USE_HIPBLAS
#define GGML_CUDA_NAME "ROCm"
#define GGML_CUBLAS_NAME "hipBLAS"
#elif defined(GGML_USE_MUSA)
#define GGML_CUDA_NAME "MUSA"
#define GGML_CUBLAS_NAME "muBLAS"
#else
#define GGML_CUDA_NAME "CUDA"
#define GGML_CUBLAS_NAME "cuBLAS"
#endif

#ifdef  __cplusplus
extern "C" {
#endif

#define GGML_CUDA_MAX_DEVICES       16

// backend API
GGML_API GGML_CALL ggml_backend_t ggml_backend_init(int device);

GGML_API GGML_CALL bool ggml_backend_is_cuda(ggml_backend_t backend);

// device buffer
GGML_API GGML_CALL ggml_backend_buffer_type_t ggml_backend_get_buffer_type(int device);

// split tensor buffer that splits matrices by rows across multiple devices
GGML_API GGML_CALL ggml_backend_buffer_type_t ggml_backend_split_buffer_type(const float * tensor_split);

// pinned host buffer for use with the CPU backend for faster copies between CPU and GPU
GGML_API GGML_CALL ggml_backend_buffer_type_t ggml_backend_host_buffer_type(void);

GGML_API GGML_CALL int  ggml_backend_get_device_count(void);
GGML_API GGML_CALL void ggml_backend_cuda_get_device_description(int device, char * description, size_t description_size);
GGML_API GGML_CALL void ggml_backend_get_device_memory(int device, size_t * free, size_t * total);

GGML_API GGML_CALL bool ggml_backend_register_host_buffer(void * buffer, size_t size);
GGML_API GGML_CALL void ggml_backend_unregister_host_buffer(void * buffer);

GGML_API void ggml_backend_log_set_callback(ggml_log_callback log_callback, void * user_data);

GGML_API enum ggml_dev_attr ggml_backend_get_dev_attr();

GGML_API size_t ggml_backend_get_max_devices(void);

GGML_API bool ggml_backend_supports_gpu_offload(void);

GGML_API bool ggml_backend_supports_pipeline_parallel(void);

GGML_CALL int ggml_backend_reg_devices();

#ifdef  __cplusplus
}
#endif
